#!/usr/bin/env python

import snmp, time
from walt.server.topology import Topology

class Platform(object):

    def __init__(self, db):
        self.db = db
        self.topology = Topology(db)

    def update(self, requester=None):
        self.topology.update(requester)

    def describe(self, details=False):
        if details:
            return self.topology.printed_as_detailed_table()
        else:
            return self.topology.printed_as_tree()

    def setpower(self, requester, node_name, poweron):
        connectivity_info = \
            self.topology.get_connectivity_info(requester, node_name)
        if connectivity_info == None:
            return False # error already reported
        # all is fine, let's do it
        switch_ip = connectivity_info['switch_ip']
        switch_port = connectivity_info['switch_port']
        proxy = snmp.Proxy(switch_ip, poe=True)
        proxy.poe.set_port(switch_port, poweron)
        return True

    def rename_device(self, requester, old_name, new_name):
        self.topology.rename_device(requester, old_name, new_name)

    def register_node(self, node_ip):
        self.topology.register_node(node_ip)

    def list_nodes(self):
        return self.topology.list_nodes()




#!/usr/bin/env python
from walt.server.snmp.vlan import PortConfig, SwitchConfig

def configure_main_switch_if_needed(snmp_proxy):
    if snmp_proxy.vlan.vlan_exists('walt-out'):
        print 'the main switch is already configured'
        return False    # already configured
    else:
        print 'configuring the main switch...'
        config = SwitchConfig()
        config.register_vlan(1, 'walt')
        config.register_vlan(169, 'walt-out')
        config.add_port_config(PortConfig(1).access(169))
        config.add_port_config(PortConfig(2).trunk([169], native_vlan_id = 1))
        snmp_proxy.vlan.apply_config_to_device(config)
        return True





COLUMNATE_SPACING = 2

# use the following like this:
#
# with AutoCleaner(<cls>) as <var>:
#     ... work_with <var> ...
#
# the <cls> must provide a method cleanup() 
# that will be called automatically when leaving 
# the with construct. 

class AutoCleaner(object):
    def __init__(self, cls):
        self.cls = cls
    def __enter__(self):
        self.instance = self.cls()
        return self.instance
    def __exit__(self, t, value, traceback):
        self.instance.cleanup()
        self.instance = None

def as_string(item):
    if item == None:
        return ''
    else:
        return str(item)

def columnate(tabular_data, header = None):
    # stringify all (and work on a copy)
    tabular_data_copy = [ [as_string(s) for s in i] for i in tabular_data ]
    # if header is specified, add it
    if header != None:
        tabular_data_copy.insert(0, header)
    # compute the max length of elements in each column
    colwidths = [ max([ len(s) for s in i ]) for i in zip(*tabular_data_copy) ]
    # if header, underline it 
    if header != None:
        tabular_data_copy.insert(1, [ '-' * w for w in colwidths ])
    # compute a format that should be applied to each record
    formating = "".join([ '%-' + str(w + COLUMNATE_SPACING) + 's' \
                            for w in colwidths ])
    # format and return
    return '\n'.join(formating % tuple(record) for record in tabular_data_copy)




#!/usr/bin/env python
import re
from snimpy import manager

class SNMPBitField(object):
    def __init__(self, snmpfield, shift = 0):
        hex_string = snmpfield.encode('hex')
        self.bitlength = len(hex_string)*4
        self.value = int(hex_string, 16)
        self.shift = shift
    def __str__(self):
        format_string = '{:0%db}' % self.bitlength
        return format_string.format(self.value)
    def bit_index(self, index):
        return self.bitlength - index - 1 + self.shift
    def __getitem__(self, index):
        bitindex = self.bit_index(index)
        return (self.value & (1 << bitindex)) >> bitindex
    def __setitem__(self, index, val):
        currval = self[index]
        if val != currval:
            self.value ^= (1 << self.bit_index(index))
        # otherwise, nothing to do.
    def toOctetString(self):
        format_string = '{:0%dx}' % (self.bitlength / 4)
        return format_string.format(self.value).decode('hex')

# For ports-related bitfields, the indexing starts at 1
# because there is no port 0.
# e.g. port 1 status is the left-most bit
# So we use the 'shift = 1' constructor option. 
class PortsBitField(SNMPBitField):
    def __init__(self, snmpfield):
        SNMPBitField.__init__(self, snmpfield, shift = 1)

def load_mib(mib):
    if not mib in manager.loaded:
        manager.load(mib)

def unload_mib(mib):
    manager.loaded.remove(mib)

def get_loaded_mibs():
    return manager.loaded

def unload_any_of_these_mibs(mibs):
    for mib in mibs:
        if mib in manager.loaded:
            unload_mib(mib)

def decode_ipv4_address(octet_string):
    return '.'.join([str(ord(i)) for i in octet_string])

def decode_mac_address(octet_string):
    return ':'.join(re.findall('..', octet_string.encode('hex')))

def enum_label(enum):
    return enum.entity.enum[enum]




import socket
from datetime import datetime
from walt.common.tcp import read_pickle, write_pickle, \
                            REQ_NEW_INCOMING_LOGS, REQ_DUMP_LOGS

class LogsToDBHandler(object):
    def __init__(self, db):
        self.db = db

    def log(self, record, stream_id):
        self.db.insert('logs',
                stream_id = stream_id,
                **record)

class LogsHub(object):
    def __init__(self):
        self.handlers = set([])

    def addHandler(self, handler):
        self.handlers.add(handler)

    def removeHandler(self, handler):
        self.handlers.remove(handler)

    def log(self, **kwargs):
        to_be_removed = set([])
        for handler in self.handlers:
            res = handler.log(**kwargs)
            # a handler may request to be deleted
            # by returning False
            if res == False:
                to_be_removed.add(handler)
        for handler in to_be_removed:
            self.handlers.remove(handler)

class LogsStreamListener(object):
    def __init__(self, db, hub, sock, sock_file, **kwargs):
        self.db = db
        self.hub = hub
        self.sock = sock
        self.sock_file = sock_file
        self.stream_id = None

    def register_new_stream(self):
        name = str(read_pickle(self.sock_file))
        sender_ip, sender_port = self.sock.getpeername()
        sender_info = self.db.select_unique('devices', ip = sender_ip)
        if sender_info == None:
            sender_mac = None
        else:
            sender_mac = sender_info.mac
        stream_id = self.db.insert('logstreams', returning='id',
                            sender_mac = sender_mac, name = name)
        # these are not needed anymore
        self.db = None
        self.sock = None
        return stream_id

    # let the event loop know what we are reading on
    def fileno(self):
        return self.sock_file.fileno()
    # when the event loop detects an event for us, we
    # know a log line should be read. 
    def handle_event(self, ts):
        if self.stream_id == None:
            self.stream_id = self.register_new_stream()
            # register_new_stream() involves a read on the stream
            # to get its name.
            # supposedly that's why we have been woken up.
            # let the event loop call us again if there is more.
            return True
        record = read_pickle(self.sock_file)
        if record == None:
            print 'Log stream with id %d is being closed.' % self.stream_id
            # let the event loop know we should 
            # be removed.
            return False
        # convert timestamp to datetime
        ts = record['timestamp']
        if not isinstance(ts, datetime):
            record['timestamp'] = datetime.fromtimestamp(ts)
        self.hub.log(record=record, stream_id=self.stream_id)
        return True
    def close(self):
        self.sock_file.close()

class LogsToSocketHandler(object):
    def __init__(self, db, hub, sock_file, **kwargs):
        self.db = db
        self.sock_file = sock_file
        self.cache = {}
        hub.addHandler(self)
    def log(self, record, stream_id):
        try:
            if stream_id not in self.cache:
                self.cache[stream_id] = self.db.execute(
                """SELECT d.name as sender, s.name as stream
                   FROM logstreams s, devices d
                   WHERE s.id = %s
                     AND s.sender_mac = d.mac
                """ % stream_id).fetchall()[0]._asdict()
            d = {}
            d.update(record)
            d.update(self.cache[stream_id])
            write_pickle(d, self.sock_file)
        except Exception as e:
            print e
            # the socket was supposedly closed.
            # notify the hub that we should be removed.
            return False
    # let the event loop know what we are reading on
    def fileno(self):
        return self.sock_file.fileno()
    # this is what we do when the event loop detects an event for us
    def handle_event(self, ts):
        return False    # no communication is expected this way
    def close(self):
        self.sock_file.close()

class LogsManager(object):
    def __init__(self, db, tcp_server):
        self.db = db
        self.hub = LogsHub()
        self.hub.addHandler(LogsToDBHandler(db))
        tcp_server.register_listener_class(
                    req_id = REQ_DUMP_LOGS,
                    cls = LogsToSocketHandler,
                    db = self.db,
                    hub = self.hub)
        tcp_server.register_listener_class(
                    req_id = REQ_NEW_INCOMING_LOGS,
                    cls = LogsStreamListener,
                    db = self.db,
                    hub = self.hub)



from plumbum.cmd import exportfs
from walt.server import const
from walt.common.tools import do, succeeds

IMAGE_EXPORT_PATTERN = """
%(image_mountpoint)s %(walt_subnet)s(fsid=%(fsid)s,rw,sync,no_root_squash,no_subtree_check)
"""

def get_fsid(image):
    return image.cid[:32]   # 32 first characters

def generate_exports_file(images):
    with open('/etc/exports', 'w') as f:
        f.write("\n".join([
            IMAGE_EXPORT_PATTERN % dict(
                image_mountpoint=image.mount_path,
                walt_subnet=const.WALT_SUBNET,
                fsid=get_fsid(image)
            ) for image in images ]))

def update_exported_filesystems(images):
    generate_exports_file(images)
    exportfs('-r')
    ensure_nfsd_is_running()

def ensure_nfsd_is_running():
    if not succeeds('pidof nfsd >/dev/null'):
        do('service nfs-kernel-server restart')



#!/usr/bin/env python
from ipaddress import ip_address, ip_network
from walt.common.tools import do, succeeds
from walt.server import const

def ip(ip_as_str):
    return ip_address(unicode(ip_as_str))

def net(net_as_str):
    return ip_network(unicode(net_as_str))

def find_free_ip_near(ip, intf, increment):
    target_ip = ip
    while True:
        target_ip += increment
        if succeeds('arping -D -w 1 -I %s %s' % (intf, target_ip)):
            return target_ip

def smallest_subnet_for_these_ip_addresses(ip1, ip2):
    # start with <ip1>/31, then <ip1>/30 etc.
    # until <ip2> is in this network too.
    for netrange in range(31,0,-1):
        net = ip_network(u'%s/%d' % (ip1, netrange), strict=False)
        if ip2 in net:
            return net

def dhcp_start(interfaces):
    do('dhclient -nw %s' % ' '.join(interfaces))

def dhcp_stop():
    do('dhclient -r')

def add_ip_to_interface(ip, subnet, intf):
    do('ip addr add %s/%d dev %s' % (ip, subnet.prefixlen, intf))

def del_ip_from_interface(ip, subnet, intf):
    do('ip addr del %s/%d dev %s' % (ip, subnet.prefixlen, intf))

def check_if_we_can_reach(remote_ip):
    return succeeds('ping -c 1 -w 1 %s' % remote_ip)

def assign_temp_ip_to_reach_neighbor(neighbor_ip, callback):
    reached = False
    callback_result = None
    for increment in [ 1, -1 ]:
        free_ip = find_free_ip_near(neighbor_ip, 'eth0', increment)
        subnet = smallest_subnet_for_these_ip_addresses(neighbor_ip, free_ip)
        add_ip_to_interface(free_ip, subnet, 'eth0')
        if check_if_we_can_reach(neighbor_ip):
            callback_result = callback(free_ip, neighbor_ip)
            reached = True
        del_ip_from_interface(free_ip, subnet, 'eth0')
        if reached:
            break
    return (reached, callback_result)

def lldp_update():
    do('lldpcli update')

def get_server_ip():
    subnet = net(const.WALT_SUBNET)
    return list(subnet.hosts()).pop(0)

def set_server_ip():
    add_ip_to_interface(
            get_server_ip(),
            net(const.WALT_SUBNET),
            'eth0')
    # let neighbors know we have updated things
    lldp_update()

def ip_in_walt_network(input_ip):
    subnet = net(const.WALT_SUBNET)
    return ip(input_ip) in subnet




#!/usr/bin/env python
from base import load_mib, decode_ipv4_address, \
                    decode_mac_address, enum_label

class LLDPProxy(object):
    def __init__(self, snmp_proxy):
        self.snmp = snmp_proxy
        load_mib("LLDP-MIB")

    def get_neighbors(self):

        mac_per_port = {}
        ip_per_port = {}

        # retrieve mac addresses of neighbors
        chassis_types = dict(self.snmp.lldpRemChassisIdSubtype)
        chassis_values = dict(self.snmp.lldpRemChassisId)
        for neighbor_key in chassis_types:
            if enum_label(chassis_types[neighbor_key]) == 'macAddress':
                timeMark, port, index = neighbor_key
                mac_per_port[int(port)] = decode_mac_address(
                                chassis_values[neighbor_key])

        # retrieve ip addresses of neighbors
        ip_info = set(self.snmp.lldpRemManAddrIfSubtype)
        for neighbor_ip_info in ip_info:
            timeMark, port, index, subtype, encoded_ip = neighbor_ip_info
            if enum_label(subtype).lower() == 'ipv4':
                ip_per_port[int(port)] = decode_ipv4_address(encoded_ip)

        # merge info
        neighbors = {}
        for port, mac in mac_per_port.items():
            ip = ip_per_port[port] if port in ip_per_port else None
            neighbors[port] = { 'mac': mac, 'ip': ip }

        return neighbors

    def get_local_ips(self):
        ips = []
        ip_info = set(self.snmp.lldpLocManAddrIfSubtype)
        for subtype, encoded_ip in ip_info:
            if enum_label(subtype).lower() == 'ipv4':
                ips.append(decode_ipv4_address(encoded_ip))
        return ips




#!/usr/bin/env python

from walt.server.const import WALT_DBNAME, WALT_DBUSER
from walt.server.tools import columnate
import psycopg2, os, shlex
from psycopg2.extras import NamedTupleCursor
from subprocess import Popen, PIPE
from sys import stdout, stderr

class PostgresDB():

    def __init__(self):
        self.conn = None
        while self.conn == None:
            try:
                self.conn = psycopg2.connect(
                        database=WALT_DBNAME, user=WALT_DBUSER)
            except psycopg2.OperationalError:
                self.create_db_and_user()
        # allow name-based access to columns
        self.c = self.conn.cursor(cursor_factory = NamedTupleCursor)

    def __del__(self):
        self.conn.commit()
        self.c.close()
        self.conn.close()

    def create_db_and_user(self):
        # we must use the postgres admin user for this
        args = shlex.split('su -c psql -l postgres')
        popen = Popen(args, stdin=PIPE, stdout=stdout, stderr=stderr)
        popen.stdin.write('''
                CREATE USER %(user)s;
                ALTER ROLE %(user)s WITH CREATEDB;
                CREATE DATABASE %(db)s OWNER %(user)s;
                ''' % dict(user=WALT_DBUSER, db=WALT_DBNAME))
        popen.stdin.close()
        popen.wait()

    def commit(self):
        self.conn.commit()

    def execute(self, query, query_args = None):
        self.c.execute(query, query_args)
        return self.c

    # from a dictionary of the form <col_name> -> <value>
    # we want to filter-out keys that are not column names,
    # and return ([<col_name>, ...], [<value>, ...]).
    def get_cols_and_values(self, table, dictionary):
        # retrieve fields names for this table 
        self.c.execute("SELECT * FROM %s LIMIT 0" % table)
        col_names = set(col_desc[0] for col_desc in self.c.description)

        res = {}
        for k in dictionary:
            # filter-out keys of dictionary that are not 
            # a column name
            if k not in col_names:
                continue
            # store in the result dict
            res[k] = dictionary[k]
        # we prefer a tuple ([k,...],[v,...]) instead of
        # a dictionary, because in an insert query,
        # we need a list of keys and a list of values 
        # with the same ordering.
        items = res.items()
        return (list(t[0] for t in items),
                list(t[1] for t in items))

    # allow statements like:
    # db.insert("network", ip=ip, switch_ip=swip)
    def insert(self, table, returning=None, **kwargs):
        # insert and return True or return False
        cols, values = self.get_cols_and_values(table, kwargs)
        sql = """INSERT INTO %s(%s)
                VALUES (%s)""" % (
                    table,
                    ','.join(cols),
                    ','.join(['%s'] * len(values)))
        if returning:
            sql += " RETURNING %s" % returning
        self.c.execute(sql + ';', tuple(values))
        if returning:
            return self.c.fetchone()[0]

    # allow statements like:
    # db.update("topology", "mac", switch_mac=swmac, switch_port=swport)
    def update(self, table, primary_key_name, **kwargs):
        cols, values = self.get_cols_and_values(table, kwargs)
        values.append(kwargs[primary_key_name])
        self.c.execute("""
                UPDATE %s 
                SET %s
                WHERE %s = %%s;""" % (
                    table,
                    ','.join("%s = %%s" % col for col in cols),
                    primary_key_name),
                    values)
        return self.c.rowcount  # number of rows updated

    # allow statements like:
    # mem_db.select("network", ip=ip)
    def select(self, table, **kwargs):
        cols, values = self.get_cols_and_values(table, kwargs)
        constraints = [ "%s=%%s" % col for col in cols ]
        if len(constraints) > 0:
            where_clause = "WHERE %s" % (
                ' AND '.join(constraints));
        else:
            where_clause = ""
        sql = "SELECT * FROM %s %s;" % (table, where_clause)
        self.c.execute(sql, values)
        return self.c.fetchall()

    # same as above but expect only one matching record
    # and return it.
    def select_unique(self, table, **kwargs):
        record_list = self.select(table, **kwargs)
        if len(record_list) == 0:
            return None
        else:
            return record_list[0]

    def pretty_printed_table(self, table):
        return self.pretty_printed_select("select * from %s;" % table)

    def pretty_printed_select(self, select_query):
        res = self.execute(select_query).fetchall()
        return columnate(res, header=res[0]._fields)




#!/usr/bin/env python

import uuid
from walt.common.evloop import EventLoop
from walt.common.tcp import TCPServer
from walt.common.constants import WALT_SERVER_TCP_PORT
from walt.server.image import NodeImageRepository
from walt.server.platform import Platform
from walt.server.db import ServerDB
from walt.server.logs import LogsManager
from walt.server.network.dhcpd import DHCPServer
from walt.server.interactive import InteractionManager

class Server(object):

    def __init__(self):
        self.ev_loop = EventLoop()
        self.db = ServerDB()
        self.platform = Platform(self.db)
        self.images = NodeImageRepository(self.db)
        self.dhcpd = DHCPServer(self.db)
        self.tcp_server = TCPServer(WALT_SERVER_TCP_PORT)
        self.logs = LogsManager(self.db, self.tcp_server)
        self.interaction = InteractionManager(\
                        self.tcp_server, self.ev_loop)
        self.tcp_server.join_event_loop(self.ev_loop)
        self.db.plan_auto_commit(self.ev_loop)

    def update(self):
        # ensure the dhcp server is running,
        # otherwise the switches may have ip addresses
        # outside the WalT network, and we will not be able
        # to communicate with them when trying to update
        # the topology.
        self.dhcpd.update(force=True)
        # topology exploration
        self.platform.topology.update()
        # update dhcp again for any new device
        self.dhcpd.update()
        # mount images needed
        self.images.update_image_mounts()

    def cleanup(self):
        self.images.cleanup()

    def set_image(self, requester, node_name, image_name):
        node_info = self.platform.topology.get_node_info(
                        requester, node_name)
        if node_info == None:
            return # error already reported
        mac = node_info.mac
        self.images.set_image(requester, mac, image_name)
        self.dhcpd.update()

    def set_default_image(self, requester, image_name):
        self.images.set_default(requester, image_name)
        self.dhcpd.update()

    def register_node(self, node_ip):
        self.platform.register_node(node_ip)
        self.dhcpd.update()

    def rename_device(self, requester, old_name, new_name):
        self.platform.rename_device(requester, old_name, new_name)
        self.dhcpd.update()

    def platform_update(self, requester):
        self.platform.update(requester)
        self.dhcpd.update()




#!/usr/bin/env python
import copy
from snimpy.manager import Manager
from poe import PoEProxy
from lldp import LLDPProxy
from vlan import VlanProxy
from walt.server import const

SNMP_OPTS = {
    "version": 2,
    "community": "private",
    'retries': 2,
    'timeout': const.SNMP_TIMEOUT
}

class Proxy(object):
        def __init__(self, host, poe=False, lldp=False, vlan=False):
            opts = SNMP_OPTS.copy()
            opts['host'] = host
            self.host = host
            self.snmp = Manager(**opts)
            if poe:
                self.poe = PoEProxy(self.snmp, host)
            if lldp:
                self.lldp = LLDPProxy(self.snmp)
            if vlan:
                self.vlan = VlanProxy(self.snmp)




#!/usr/bin/env python
import os, pty, shlex, uuid, fcntl, termios, sys
from walt.common.io import SmartBufferingFileReader, \
                            unbuffered, read_and_copy
from walt.common.tcp import REQ_SQL_PROMPT, REQ_DOCKER_PROMPT, \
                            REQ_NODE_SHELL, REQ_DEVICE_PING, \
                            read_pickle

class PromptProcessListener(object):
    def __init__(self, slave_r, slave_w, sock_file_r, sock_file_w):
        self.slave_r = slave_r
        self.slave_w = slave_w
        self.sock_file_r = sock_file_r
        self.sock_file_w = sock_file_w
        self.slave_reader = SmartBufferingFileReader(slave_r)
    # let the event loop know what we are reading on
    def fileno(self):
        return self.slave_r.fileno()
    # when the event loop detects an event for us, this means
    # the slave proces wrote something on its output
    # we just have to copy this to the user socket
    def handle_event(self, ts):
        return read_and_copy(
                self.slave_reader, self.sock_file_w)
    def close(self):
        # note: if we close, we want the other listener
        # (PromptSocketListener) to close too.
        # that's why we kept a reference to sock_file_r
        # and slave_w, closing them should achieve what
        # we want.
        self.sock_file_r.close()
        self.sock_file_w.close()
        self.slave_r.close()
        self.slave_w.close()

class PromptSocketListener(object):
    def __init__(self, ev_loop, sock_file, **kwargs):
        self.ev_loop = ev_loop
        self.win_size = None
        self.cmd = None
        # use unbuffered reading & writing on the socket
        self.sock_file_r = unbuffered(sock_file, 'r')
        self.sock_file_w = unbuffered(sock_file, 'w')
        self.slave_r = None
        self.slave_w = None
    def get_command(self):
        '''this should be defined in subclasses'''
        raise NotImplementedError
    def start(self):
        self.slave_pid, fd_slave = pty.fork()
        # the child (slave process) should execute the command
        if self.slave_pid == 0:
            # set the window size appropriate for the remote client
            fcntl.ioctl(sys.stdout.fileno(), termios.TIOCSWINSZ, self.win_size)
            env = os.environ.copy()
            env.update({'TERM':'xterm'})
            cmd_args = shlex.split(self.cmd)
            os.execvpe(cmd_args[0], cmd_args, env)
        # the parent should communicate.
        # use unbuffered communication with the slave process
        self.slave_r = os.fdopen(os.dup(fd_slave), 'r', 0)
        self.slave_w = os.fdopen(os.dup(fd_slave), 'w', 0)
        # provide read_available() method on the socket
        self.sock_reader = SmartBufferingFileReader(self.sock_file_r)
        # create a new listener on the event loop for reading
        # what the slave process outputs
        process_listener = PromptProcessListener(
                    self.slave_r, self.slave_w,
                    self.sock_file_r, self.sock_file_w)
        self.ev_loop.register_listener(process_listener)
    # let the event loop know what we are reading on
    def fileno(self):
        if self.sock_file_r.closed:
            return None
        return self.sock_file_r.fileno()
    # handle_event() will be called when the event loop detects
    # new input data for us.
    def handle_event(self, ts):
        if self.win_size == None:
            # we did not read the window size yet, let's do it
            self.win_size = read_pickle(self.sock_file_r)
            if self.win_size == None:
                self.close()    # issue
        elif self.cmd == None:
            # we did not compute the command to be run yet, let's do it
            self.cmd = self.get_command()
            if self.cmd == None:
                self.close()    # issue
                return
            # we now have all info to start the child process
            self.start()
        else:
            # otherwise we are all set. Thus, getting input data means
            # the user wrote something on the prompt (i.e. the socket)
            # we just have to copy this to the slave process input
            return read_and_copy(
                self.sock_reader, self.slave_w)
    def close(self):
        self.sock_file_r.close()
        self.sock_file_w.close()
        if self.slave_r:
            self.slave_r.close()
        if self.slave_w:
            self.slave_w.close()

class SQLPromptSocketListener(PromptSocketListener):
    REQ_ID = REQ_SQL_PROMPT
    def get_command(self):
        return 'psql walt'

class DockerPromptSocketListener(PromptSocketListener):
    REQ_ID = REQ_DOCKER_PROMPT
    def get_command(self):
        image, container = read_pickle(self.sock_file_r)
        return 'docker run -it --entrypoint %s -h %s --name %s %s' % \
                       ('/bin/bash', 'image-modify',
                        container, image)

class NodeShellSocketListener(PromptSocketListener):
    REQ_ID = REQ_NODE_SHELL
    def get_command(self):
        node_ip = read_pickle(self.sock_file_r)
        return 'ssh -o StrictHostKeyChecking=no root@%s' % node_ip

class DevicePingSocketListener(PromptSocketListener):
    REQ_ID = REQ_DEVICE_PING
    def get_command(self):
        device_ip = read_pickle(self.sock_file_r)
        return 'ping %s' % device_ip

class InteractionManager(object):
    def __init__(self, tcp_server, ev_loop):
        for cls in [    SQLPromptSocketListener,
                        DockerPromptSocketListener,
                        NodeShellSocketListener,
                        DevicePingSocketListener ]:
            tcp_server.register_listener_class(
                    req_id = cls.REQ_ID,
                    cls = cls,
                    ev_loop = ev_loop)




#!/usr/bin/env python

from collections import OrderedDict

# special characters used to display the tree
WHOLE_V = u'\u2502'         # whole vertical bar
WHOLE_V_RIGHT_H = u'\u251c' # whole vertical + right horizontal bars
WHOLE_H = u'\u2500'         # whole horizontal bar
UPPER_V_RIGHT_H = u'\u2514' # upper vertical + right horizontal bars
SPACE = u' '

class Tree(object):
    """Class allowing to display an object hierarchy as a tree."""
    def __init__(self):
        self.nodes = OrderedDict()
        self.up_to_date = False
        self.root = None
    def add_node(self, key, label, subtree_offset=0, parent_key = None):
        self.nodes[key] = dict(
            key=key,
            label=label,
            subtree_offset=subtree_offset,
            parent_key=parent_key)
        if parent_key == None:
            self.root = self.nodes[key]
        self.up_to_date = False
    def compute_children(self):
        for node in self.nodes.values():
            node['children'] = []
        for node in self.nodes.values():
            parent_key = node['parent_key']
            if parent_key != None:
                node_key = node['key']
                parent = self.nodes[parent_key]
                parent['children'].append(node_key)
    def printed(self):
        if self.up_to_date == False:
            self.compute_children()
            self.up_to_date = True
        return self.print_elem(**self.root)
    def print_elem(self, label, subtree_offset, parent_key, \
                        children, prefix = '', \
                        last_child = False, **kwargs):
        if parent_key == None:  # root element
            output = "%s\n" % label
            prefix += (SPACE * subtree_offset)
        else:
            if last_child:
                sep_char_item = UPPER_V_RIGHT_H
                sep_char_child = SPACE
            else:
                sep_char_item = WHOLE_V_RIGHT_H
                sep_char_child = WHOLE_V
            output = "%s%s%s%s\n" % \
                    (prefix, sep_char_item, WHOLE_H, label)
            prefix += sep_char_child + (SPACE * (subtree_offset+1))
        num_children = len(children)
        for idx, child_key in enumerate(children):
                last_child = (idx == num_children-1)
                child = self.nodes[child_key]
                output += self.print_elem(
                                prefix = prefix,
                                last_child = last_child,
                                **child)
        return output




#!/usr/bin/env python

import rpyc, sys
import walt.server as server
from walt.server.network import setup
from walt.server.tools import AutoCleaner
from walt.common.daemon import WalTDaemon
from walt.common.constants import           \
                 WALT_SERVER_DAEMON_PORT,   \
                 WALT_NODE_DAEMON_PORT

WALT_SERVER_DAEMON_VERSION = 0.1

class ClientMirroringService(rpyc.Service):
    services_per_node = {}

    def on_connect(self):
        node_id = id(self._conn.root)
        self._node_id = node_id
        ClientMirroringService.services_per_node[node_id] = self

    def on_disconnect(self):
        del ClientMirroringService.services_per_node[self._node_id]

    def __del__(self):
        self._client = None

    def register_client(self, client):
        self._client = client

    @staticmethod
    def link_node_to_client(node, client):
        service = ClientMirroringService.services_per_node[id(node)]
        service.register_client(client)

    # forward all other method accesses to self._client
    def __getattr__(self, attr_name):
        return getattr(self._client, attr_name)

class ServerToNodeLink:
    def __init__(self, ip_address, client = None):
        self.node_ip = ip_address
        self.client = client

    def __enter__(self):
        if self.client:
            self.conn = rpyc.connect(self.node_ip, WALT_NODE_DAEMON_PORT,
                            service = ClientMirroringService)
            node_service = self.conn.root
            ClientMirroringService.link_node_to_client(node_service, self.client)
        else:
            self.conn = rpyc.connect(self.node_ip, WALT_NODE_DAEMON_PORT)
        return self.conn.root

    def __exit__(self, type, value, traceback):
        self.conn.close()

class PlatformService(rpyc.Service):
    ALIASES=("WalT_Platform",)

    def __init__(self, *args, **kwargs):
        rpyc.Service.__init__(self, *args, **kwargs)
        self.platform = server.instance.platform
        self.images = server.instance.images
        self.server = server.instance

    def on_connect(self):
        self._client = self._conn.root

    def on_disconnect(self):
        self._client = None

    def exposed_update(self):
        self.server.platform_update(self._client)

    def exposed_describe(self, details=False):
        return self.platform.describe(details)

    def exposed_list_nodes(self):
        return self.platform.list_nodes()

    def exposed_get_reachable_node_ip(self, node_name):
        return self.platform.topology.get_reachable_node_ip(
                        self._client, node_name)

    def exposed_get_device_ip(self, device_name):
        return self.platform.topology.get_device_ip(
                        self._client, device_name)

    def exposed_get_node_ip(self, node_name):
        return self.platform.topology.get_node_ip(
                        self._client, node_name)

    def exposed_blink(self, node_name, blink_status):
        node_ip = self.platform.topology.get_reachable_node_ip(
                        self._client, node_name)
        if node_ip == None:
            return False # error was already reported
        with ServerToNodeLink(node_ip, self._client) as node_service:
            node_service.blink(blink_status)
        return True

    def exposed_poweroff(self, node_name):
        return self.platform.setpower(self._client, node_name, False)

    def exposed_poweron(self, node_name):
        return self.platform.setpower(self._client, node_name, True)

    def exposed_rename(self, old_name, new_name):
        self.server.rename_device(self._client, old_name, new_name)

    def exposed_register_node(self):
        node_ip, node_port = self._conn._config['endpoints'][1]
        self.server.register_node(node_ip)

    def exposed_has_image(self, image_name):
        return self.images.has_image(self._client, image_name)

    def exposed_set_image(self, node_name, image_name):
        self.server.set_image(self._client, node_name, image_name)

    def exposed_list_images(self):
        return self.images.describe()

    def exposed_set_default_image(self, image_name):
        self.server.set_default_image(self._client, image_name)

    def exposed_create_modify_image_session(self, image_name):
        return self.images.create_modify_session(self._client, image_name)

    def exposed_remove_image(self, image_name):
        self.images.remove(self._client, image_name)

    def exposed_rename_image(self, image_name, new_name):
        self.images.rename(self._client, image_name, new_name)

class WalTServerDaemon(WalTDaemon):
    """WalT (wireless testbed) server daemon."""
    VERSION = WALT_SERVER_DAEMON_VERSION

    def getParameters(self):
        return dict(
                service_cl = PlatformService,
                port = WALT_SERVER_DAEMON_PORT,
                ev_loop = server.instance.ev_loop)

def run():
    if setup.setup_needed():
        setup.setup()
    with AutoCleaner(server.Server) as server.instance:
        server.instance.update()
        WalTServerDaemon.run()

if __name__ == "__main__":
    run()




#!/usr/bin/env python
from base import load_mib, PortsBitField

Q_BRIDGE_MIB_CreateAndGo = 4

class PortConfig(object):
    def __init__(self, port):
        self.port = port
        self.default_vlan_id = 1
        self.egress_vlans = set([ 1 ])
        self.untagged_vlans = set([ 1 ])
    def trunk(self, vlan_id_set, native_vlan_id = None):
        self.egress_vlans = set(vlan_id_set)
        self.untagged_vlans = set([])
        if native_vlan_id != None:
            self.egress_vlans.add(native_vlan_id)
            self.untagged_vlans.add(native_vlan_id)
            self.default_vlan_id = native_vlan_id
        return self
    def access(self, vlan_id):
        self.default_vlan_id = vlan_id
        self.egress_vlans = set([vlan_id])
        self.untagged_vlans = set([vlan_id])
        return self

class SwitchConfig(object):
    def __init__(self):
        self.port_configs = set([])
        self.registered_vlans = {}
    def register_vlan(self, vlan_id, vlan_name):
        self.registered_vlans[vlan_id] = vlan_name
    def add_port_config(self, port_config):
        self.port_configs.add(port_config)

class VlanProxy(object):
    def __init__(self, snmp_proxy):
        load_mib("Q-BRIDGE-MIB")
        self.snmp = snmp_proxy
    def get_existing_vlans(self):
        static_vlans = dict(self.snmp.dot1qVlanStaticName)
        existing_vlans = {}
        for vlan_id in [ int(k[1]) for k in self.snmp.dot1qVlanStatus.keys() ]:
            if vlan_id in static_vlans:
                existing_vlans[vlan_id] = { 'preconfigured': False,
                                            'name': str(static_vlans[vlan_id]) }
            else:
                existing_vlans[vlan_id] = { 'preconfigured': True }
        return existing_vlans
    def vlan_exists(self, vlan_name):
        return vlan_name in self.snmp.dot1qVlanStaticName.values()
    def apply_config_to_device(self, config):
        print 'setting the default vlan id on each port...'
        for port_config in config.port_configs:
            self.snmp.dot1qPvid[port_config.port] = port_config.default_vlan_id
        print 'configuring vlans...'
        existing_vlans = self.get_existing_vlans()
        for vlan_id in config.registered_vlans:
            print 'vlan', vlan_id
            # create missing vlans and update vlan names.
            # note: this operation will update the size of the 2 tables used below
            # dot1qVlanStaticEgressPorts & dot1qVlanStaticUntaggedPorts
            if not vlan_id in existing_vlans:
                print '-> new vlan, creating it...'
                self.snmp.dot1qVlanStaticRowStatus[vlan_id] = Q_BRIDGE_MIB_CreateAndGo
                existing_vlans[vlan_id] = { 'preconfigured': False,
                                            'name': '' }
            if existing_vlans[vlan_id]['preconfigured']:
                print '-> bypassing vlan name update (this vlan was preconfigured in the device)'
            else:
                new_vlan_name = config.registered_vlans[vlan_id]
                old_name = existing_vlans[vlan_id]['name']
                if old_name != new_vlan_name:
                    print '-> setting vlan name to', new_vlan_name, '(was %s)' % old_name
                    self.snmp.dot1qVlanStaticName[vlan_id] = new_vlan_name
                print '-> retrieving current config egress / untagged ports'
                egress_ports = PortsBitField(self.snmp.dot1qVlanStaticEgressPorts[vlan_id])
                untagged_ports = PortsBitField(self.snmp.dot1qVlanStaticUntaggedPorts[vlan_id])
                print '-> applying changes'
                for port_config in config.port_configs:
                    egress = 1 if vlan_id in port_config.egress_vlans else 0
                    untag = 1 if vlan_id in port_config.untagged_vlans else 0
                    egress_ports[port_config.port] = egress
                    untagged_ports[port_config.port] = untag
                print '-> saving egress / untagged ports config on device'
                self.snmp.dot1qVlanStaticEgressPorts[vlan_id] = egress_ports.toOctetString()
                self.snmp.dot1qVlanStaticUntaggedPorts[vlan_id] = untagged_ports.toOctetString()




#!/usr/bin/env python
import time, snimpy
from walt.server import snmp, status, network
from waltvlanconf import configure_main_switch_if_needed

def main_switch_conf_callback(local_ip, switch_ip):
    print 'with %s we can reach the switch' % local_ip
    status.update('ensuring proper vlan configuration of the main switch...')
    configured = configure_main_switch_if_needed(
                        snmp.Proxy(str(switch_ip), vlan=True))
    return configured

def setup_needed():
    # check if an ip was already assigned, which 
    # shows that setup was already done
    server_ip = str(network.tools.get_server_ip())
    while True:
        try:
            snmp_local = snmp.Proxy('localhost', lldp=True)
            if server_ip in snmp_local.lldp.get_local_ips():
                return False    # setup already done
            else:
                return True     # setup needed
        except snimpy.snmp.SNMPNoSuchObject:
            print 'Not ready, waiting...'
            time.sleep(1)
        else:
            break

def setup():
    # unable connectivity for walt client to be able
    # to get our status through snmp 
    network.tools.dhcp_start(['eth0', 'eth0.169'])
    status.update('detecting the main switch...')
    # retrieve info about the main switch by
    # using the local lldp & snmp daemons
    snmp_local = snmp.Proxy('localhost', lldp=True)
    while len(snmp_local.lldp.get_neighbors()) == 0:
        time.sleep(0.5)
    main_switch_info = snmp_local.lldp.get_neighbors().values()[0]
    print 'main switch:', main_switch_info
    # assign a temporary ip in order to communicate
    # with the main switch, and, if not done yet, set up
    # its vlan configuration
    status.update('detecting parameters for communication with the main switch...')
    switch_ip = network.tools.ip(main_switch_info['ip'])
    reached, configured = network.tools.assign_temp_ip_to_reach_neighbor(
                                switch_ip,
                                main_switch_conf_callback)
    if reached == False:
        status.update('ERROR: unable to communicate with the main switch (try to reset it to factory settings).')
        return False    # not ok
    # set up the server ip on eth0 (walt testbed network)
    # and restart the dhcp client on eth0.169 only (walt-out vlan)
    network.tools.dhcp_stop()
    network.tools.set_server_ip()
    network.tools.dhcp_start(['eth0.169'])
    return True     # ok





SERVER_TESTBED_INTERFACE    = "eth0"
WALT_SUBNET                 = "192.168.152.0/24"
DEFAULT_IMAGE               = 'default'
SNMP_TIMEOUT                = 20
DOCKER_HUB_GET_TAGS_URL     = 'https://registry.hub.docker.com/v1/repositories/%(image_name)s/tags'
WALT_DBNAME                 = "walt"
WALT_DBUSER                 = "root"



#!/usr/bin/env python
from base import load_mib, unload_mib, get_loaded_mibs, \
                    unload_any_of_these_mibs

POE_PORT_ENABLED=1
POE_PORT_DISABLED=2

# Several mibs exist depending on the device vendor.
# In the worst case, on the WalT platform we
# could find devices of several vendors. 
# In this case we should unload a mib and load another
# when we pass from one vendor to another.
# In order to handle this, we use a wrapper around the
# SNMP proxy. This wrapper ensures that each time the SNMP
# proxy is accessed, the appropriate MIB is loaded.

CANDIDATE_POE_MIBS = ["POWER-ETHERNET-MIB", "NETGEAR-POWER-ETHERNET-MIB"]

def detect_correct_mib(snmp_proxy, host):
    unload_any_of_these_mibs(CANDIDATE_POE_MIBS)
    for mib in CANDIDATE_POE_MIBS:
        load_mib(mib)
        try:
            dummy = snmp_proxy.pethPsePortAdminEnable.keys()
            break # ok previous line passed with no error
        except:
            unload_mib(mib)
            mib = None
    if mib == None:
        raise RuntimeError(
            'Device %s does not seem to handle PoE SNMP requests.' % \
                        host)
    return mib

def ensure_correct_mib(correct_mib):
    if correct_mib not in get_loaded_mibs():
        unload_any_of_these_mibs(CANDIDATE_POE_MIBS)
        load_mib(correct_mib)

class SafeProxy(object):
    def __init__(self, snmp_proxy, host):
        # avoid the modified __setattr__ (see below) to be called
        object.__setattr__(self, 'unsafe_proxy', snmp_proxy)
        object.__setattr__(self, 'selected_mib',
                        detect_correct_mib(snmp_proxy, host))

    # The following __getattr__ and __setattr__ functions
    # ensure that before accessing the SNMP proxy,
    # the PoE MIB appropriate for this device is loaded.
    def __getattr__(self, name):
        ensure_correct_mib(self.selected_mib)
        return getattr(self.unsafe_proxy, name)

    def __setattr__(self, name, value):
        ensure_correct_mib(self.selected_mib)
        self.unsafe_proxy.__setattr__(name, value)

class PoEProxy(object):

    def __init__(self, snmp_proxy, host):
        self.snmp = SafeProxy(snmp_proxy, host)

    def set_port(self, switch_port, active_or_not):
        port_state = POE_PORT_ENABLED if active_or_not else POE_PORT_DISABLED
        self.snmp.pethPsePortAdminEnable[(1,switch_port)] = port_state




from docker import Client
from plumbum.cmd import mount, umount, findmnt
from network import nfs
from network.tools import get_server_ip
from walt.server.tools import columnate
from walt.common.tools import \
        failsafe_makedirs, failsafe_symlink
from walt.server import const
import os, re, sys, requests, uuid, shlex
from datetime import datetime

IMAGE_IS_USED_BUT_NOT_FOUND=\
    "WARNING: image %s is not found. Cannot attach it to related nodes.\n"
IMAGE_MOUNT_PATH='/var/lib/walt/images/%s/fs'
CONFIG_ITEM_DEFAULT_IMAGE='default_image'
SERVER_PUBKEY_PATH = '/root/.ssh/id_dsa.pub'
HOSTS_FILE_CONTENT="""\
127.0.0.1   localhost
::1     localhost ip6-localhost ip6-loopback
ff02::1     ip6-allnodes
ff02::2     ip6-allrouters
"""

def get_mount_path(image_name):
    return IMAGE_MOUNT_PATH % \
            re.sub('[^a-zA-Z0-9/-]+', '_', re.sub(':', '/', image_name))

def get_server_pubkey():
    with open(SERVER_PUBKEY_PATH, 'r') as f:
        return f.read()

class ModifySession(object):
    def __init__(self, requester, docker_image_name, image_name, repo):
        self.requester = requester
        self.docker_image_name = docker_image_name
        self.image_name = image_name
        self.new_image_name = None
        self.repo = repo
        self.container_name = str(uuid.uuid4())
        self.repo.register_modify_session(self)
        # expose methods to the RPyC client
        self.exposed___enter__ = self.__enter__
        self.exposed___exit__ = self.__exit__
        self.exposed_get_parameters = self.get_parameters
        self.exposed_get_default_new_name = self.get_default_new_name
        self.exposed_validate_new_name = self.validate_new_name
        self.exposed_select_new_name = self.select_new_name
    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        self.repo.finalize_modify(self)
    def get_parameters(self):
        # return an immutable object (a tuple, not a dict)
        # otherwise we will cause other RPyC calls
        return self.docker_image_name, self.container_name
    def get_default_new_name(self):
        return self.repo.get_default_new_image_name(
            self.image_name
        )
    def validate_new_name(self, new_image_name):
        return self.repo.validate_new_image_name(
            self.requester,
            new_image_name
        )
    def select_new_name(self, new_image_name):
        self.new_image_name = new_image_name

class NodeImage(object):
    server_pubkey = get_server_pubkey()
    REMOTE = 0
    LOCAL = 1
    def __init__(self, c, name, state):
        self.c = c
        self.rename(name)
        self.set_created_at()
        self.state = state
        self.cid = None
        self.mount_path = None
        self.mounted = False
        self.server_ip = get_server_ip()
    def rename(self, name):
        parts = name.split(':')
        if len(parts) == 1:
            self.name, self.tag = name, 'latest'
        else:
            self.name, self.tag = parts
        self.docker_user = name.split('/')[0]
        self.tagged_name = name
    def set_created_at(self):
        # created_at is only available on local images
        # (downloaded or created locally)
        self.created_at = None
        for i in self.c.images():
            if self.tagged_name in i['RepoTags']:
                self.created_at = datetime.fromtimestamp(i['Created'])
    def __del__(self):
        if self.mounted:
            self.unmount()
    def download(self):
        for idx, line in enumerate(
                    self.c.pull(
                        self.name,
                        tag=requests.utils.quote(self.tag),
                        stream=True)):
            progress = "-/|\\"[idx % 4]
            sys.stdout.write('Downloading image %s... %s\r' % \
                                (self.tagged_name, progress))
            sys.stdout.flush()
        sys.stdout.write('\n')
        self.state = NodeImage.LOCAL
    def get_mount_path(self):
        return get_mount_path(self.tagged_name)
    def docker_command_split(self, cmd):
        args = shlex.split(cmd)
        return dict(
            entrypoint=args[0],
            command=args[1:]
        )
    def mount(self):
        print 'Mounting %s...' % self.tagged_name,
        self.mount_path = self.get_mount_path()
        failsafe_makedirs(self.mount_path)
        if self.state == NodeImage.REMOTE:
            self.download()
        params = dict(image=self.tagged_name)
        params.update(self.docker_command_split(
            'walt-node-install %s "%s"' % \
                    (self.server_ip, NodeImage.server_pubkey)))
        self.cid = self.c.create_container(**params).get('Id')
        self.c.start(container=self.cid)
        self.bind_mount()
        self.create_hosts_file()
        self.mounted = True
        print 'done'
    def unmount(self):
        print 'Un-mounting %s...' % self.tagged_name,
        umount('-lf', self.mount_path)
        self.c.kill(container=self.cid)
        self.c.wait(container=self.cid)
        self.c.remove_container(container=self.cid)
        os.rmdir(self.mount_path)
        self.mounted = False
        print 'done'
    def create_hosts_file(self):
        # since file /etc/hosts is managed by docker,
        # it appears empty on the bind mount.
        # let's create it appropriately.
        with open(self.mount_path + '/etc/hosts', 'w') as f:
            f.write(HOSTS_FILE_CONTENT)
    def list_mountpoints(self):
        return findmnt('-nlo', 'TARGET').splitlines()
    def get_mount_point(self):
        return [ line for line in self.list_mountpoints() \
                        if line.find(self.cid) != -1 ][0]
    def bind_mount(self):
        mount('-o', 'bind', self.get_mount_point(), self.mount_path)

class NodeImageRepository(object):
    def __init__(self, db):
        self.db = db
        self.c = Client(base_url='unix://var/run/docker.sock', version='auto')
        self.images = {}
        self.modify_sessions = set()
        self.add_local_images()
        self.add_remote_images()
    def add_local_images(self):
        local_images = sum([ i['RepoTags'] for i in self.c.images() ], [])
        for name in local_images:
            if '/walt-node' in name:
                self.images[name] = NodeImage(self.c, name, NodeImage.LOCAL)
    def lookup_remote_tags(self, image_name):
        url = const.DOCKER_HUB_GET_TAGS_URL % dict(image_name = image_name)
        r = requests.get(url)
        for elem in requests.get(url).json():
            tag = requests.utils.unquote(elem['name'])
            yield "%s:%s" % (image_name, tag)
    def add_remote_images(self):
        current_names = set(self.images.keys())
        remote_names = set([])
        for result in self.c.search(term='walt-node'):
            if '/walt-node' in result['name']:
                for tagged_image in self.lookup_remote_tags(result['name']):
                    remote_names.add(tagged_image)
        for name in (remote_names - current_names):
            self.images[name] = NodeImage(self.c, name, NodeImage.REMOTE)
    def __getitem__(self, key):
        return self.images[key]
    def __iter__(self):
        return self.images.iterkeys()
    def update_image_mounts(self):
        images_in_use = self.get_images_in_use()
        images_found = []
        # ensure all needed images are mounted
        for name in images_in_use:
            if name in self.images:
                img = self.images[name]
                if not img.mounted:
                    img.mount()
                images_found.append(img)
            else:
                sys.stderr.write(IMAGE_IS_USED_BUT_NOT_FOUND % name)
        # update default image link
        self.update_default_link()
        # update nfs configuration
        nfs.update_exported_filesystems(images_found)
        # unmount images that are not needed anymore
        for name in self.images:
            if name not in images_in_use:
                img = self.images[name]
                if img.mounted:
                    img.unmount()
    def cleanup(self):
        # give up modify sessions
        for session in self.modify_sessions.copy():
            self.cleanup_modify_session(session)
        # release nfs mounts
        nfs.update_exported_filesystems([])
        # unmount images
        for name in self.images:
            img = self.images[name]
            if img.mounted:
                img.unmount()
    def get_images_in_use(self):
        res = set([ item.image for item in \
            self.db.execute("""
                SELECT DISTINCT image FROM nodes""").fetchall()])
        res.add(self.get_default_image())
        return res
    def get_default_image(self):
        if len(self.images) > 0:
            default_if_not_specified = self.images.keys()[0]
        else:
            default_if_not_specified = None
        return self.db.get_config(
                    CONFIG_ITEM_DEFAULT_IMAGE,
                    default_if_not_specified)
    def update_default_link(self):
        default_image = self.get_default_image()
        default_mount_path = get_mount_path(default_image)
        default_simlink = get_mount_path('default')
        failsafe_makedirs(default_mount_path)
        failsafe_symlink(default_mount_path, default_simlink)
    def get_image_from_tag(self, image_tag, requester = None):
        for image in self.images.values():
            if image.tag == image_tag:
                return image
        if requester:
            requester.stderr.write(
                "No such image '%s'. (tip: walt image list)\n" % image_tag)
        return None
    def has_image(self, requester, image_tag):
        return self.get_image_from_tag(image_tag, requester) != None
    def set_image(self, requester, node_mac, image_tag):
        image = self.get_image_from_tag(image_tag, requester)
        if image:
            self.db.update('nodes', 'mac',
                    mac=node_mac,
                    image=image.tagged_name)
            self.update_image_mounts()
            self.db.commit()
    def describe(self):
        tabular_data = []
        header = [ 'Name', 'Origin', 'Mounted', 'Default', 'Created' ]
        default = self.get_default_image()
        missing_created_at = False
        for name, image in self.images.iteritems():
            created_at = image.created_at
            if not created_at:
                created_at = 'N/A (1)'
                missing_created_at = True
            tabular_data.append([
                        image.tag,
                        image.docker_user,
                        str(image.mounted),
                        '*' if name == default else '',
                        created_at])
        res = columnate(tabular_data, header)
        if missing_created_at:
            res += '\n\n(1): no info about creation date for remote images.'
        return res
    def set_default(self, requester, image_tag):
        image = self.get_image_from_tag(image_tag, requester)
        if image:
            self.db.set_config(
                    CONFIG_ITEM_DEFAULT_IMAGE,
                    image.tagged_name)
            self.update_image_mounts()
            self.db.commit()
    def create_modify_session(self, requester, image_tag):
        image = self.get_image_from_tag(image_tag, requester)
        if not image:
            requester.stderr.write('No such image.\n')
            return None
        else:
            return ModifySession(requester, image.tagged_name, image_tag, self)
    def get_default_new_image_name(self, old_image_tag):
        new_tag = old_image_tag
        while self.get_image_from_tag(new_tag):
            new_tag += '_new'
        return new_tag
    def validate_new_image_name(self, requester, new_image_tag):
        if self.get_image_from_tag(new_image_tag):
            requester.stderr.write('Bad name: Image already exists.\n')
            return False
        if not re.match('^[a-zA-Z0-9\-_]+$', new_image_tag):
            requester.stderr.write(\
                'Bad name: Only alnum, dash(-) and underscore(_) characters are allowed.\n')
            return False
        return True
    def register_modify_session(self, session):
        self.modify_sessions.add(session)
    def cleanup_modify_session(self, session):
        cname = session.container_name
        # kill/remove the container if it ever existed
        try:
            self.c.kill(container=cname)
        except:
            pass
        try:
            self.c.wait(container=cname)
        except:
            pass
        try:
            self.c.remove_container(container=cname)
        except:
            pass
        self.modify_sessions.remove(session)
    def finalize_modify(self, session):
        requester = session.requester
        new_image_tag = session.new_image_name
        container_name = session.container_name
        if new_image_tag:
            self.c.commit(
                    container=container_name,
                    repository='local/walt-node',
                    tag=new_image_tag,
                    message='Image modified using walt image modify')
            full_name = 'local/walt-node:%s' % new_image_tag
            self.images[full_name] = NodeImage(self.c, full_name, NodeImage.LOCAL)
            requester.stdout.write('New image %s saved.\n' % new_image_tag)
        self.cleanup_modify_session(session)
    def get_local_unmounted_image_from_tag(self, image_tag, requester):
        image = self.get_image_from_tag(image_tag, requester)
        if image:   # otherwise issue is already reported
            if image.docker_user != 'local':
                requester.stderr.write('Sorry, this operation is allowed on local images only.\n')
                return None
            if image.mounted:
                requester.stderr.write('Sorry, cannot proceed because the image is mounted.\n')
                return None
        return image

    def remove(self, requester, image_tag):
        image = self.get_local_unmounted_image_from_tag(image_tag, requester)
        if image:   # otherwise issue is already reported
            name = image.tagged_name
            del self.images[name]
            self.c.remove_image(
                    image=name, force=True)

    def rename(self, requester, image_tag, new_tag):
        image = self.get_local_unmounted_image_from_tag(image_tag, requester)
        if image:   # otherwise issue is already reported
            if self.get_image_from_tag(new_tag):
                requester.stderr.write('Bad name: Image already exists.\n')
                return
            name = image.tagged_name
            new_name = 'local/walt-node:' + new_tag
            # update image internal attributes
            image.rename(new_name)
            # rename in this repo
            self.images[new_name] = image
            del self.images[name]
            # rename the docker image
            self.c.tag(image=name, repository='local/walt-node', tag=new_tag)
            self.c.remove_image(image=name, force=True)



#!/usr/bin/env python
from walt.server.postgres import PostgresDB
from time import time

EV_AUTO_COMMIT              = 0
EV_AUTO_COMMIT_PERIOD       = 2

class ServerDB(PostgresDB):

    def __init__(self):
        # parent constructor
        PostgresDB.__init__(self)
        # create the db schema
        self.execute("""CREATE TABLE IF NOT EXISTS devices (
                    mac TEXT PRIMARY KEY,
                    ip TEXT,
                    name TEXT,
                    reachable INTEGER,
                    type TEXT);""")
        self.execute("""CREATE TABLE IF NOT EXISTS topology (
                    mac TEXT REFERENCES devices(mac),
                    switch_mac TEXT REFERENCES devices(mac),
                    switch_port INTEGER);""")
        self.execute("""CREATE TABLE IF NOT EXISTS nodes (
                    mac TEXT REFERENCES devices(mac),
                    image TEXT);""")
        self.execute("""CREATE TABLE IF NOT EXISTS config (
                    item TEXT PRIMARY KEY,
                    value TEXT);""")
        self.execute("""CREATE TABLE IF NOT EXISTS logstreams (
                    id SERIAL PRIMARY KEY,
                    sender_mac TEXT REFERENCES devices(mac),
                    name TEXT);""")
        self.execute("""CREATE TABLE IF NOT EXISTS logs (
                    stream_id INTEGER REFERENCES logstreams(id),
                    timestamp TIMESTAMP,
                    line TEXT);""")

    # Some types of events are numerous and commiting the
    # database each time would be costly.
    # That's why we auto-commit every few seconds.
    def plan_auto_commit(self, ev_loop):
        ev_loop.plan_event(
            ts = time(),
            target = self,
            repeat_delay = EV_AUTO_COMMIT_PERIOD,
            ev_type = EV_AUTO_COMMIT
        )

    def handle_planned_event(self, ev_type):
        assert(ev_type == EV_AUTO_COMMIT)
        self.commit()

    def get_config(self, item, default = None):
        res = self.select_unique("config", item=item)
        if res == None:
            if default == None:
                raise RuntimeError(\
                    "Failed get_config(): item not found and no default provided.")
            self.insert("config", item=item, value=default)
            self.commit()
            return default
        else:
            return res.value

    def set_config(self, item, value):
        res = self.select_unique("config", item=item)
        if res == None:
            self.insert("config", item=item, value=value)
        else:
            self.update("config", "item", item=item, value=value)



#!/usr/bin/env python
from walt.server import const
from walt.server.network.tools import ip, net
from walt.common.nodetypes import is_a_node_type_name
from operator import itemgetter
from itertools import groupby
from walt.server import image
from walt.common.tools import do

DHCPD_CONF_FILE = '/etc/dhcp/dhcpd.conf'

CONF_PATTERN = """
#
# DO NOT EDIT THIS FILE
#
# It is automatically generated by the walt system 
# and updated when needed.
#

# allow sending DHCP NAKs
authoritative;

# custom parameters declaration
option nfs-server code 138 = string;
option nfs-fs-path code 140 = string;
if exists dhcp-parameter-request-list {
    option dhcp-parameter-request-list=concat(
           option dhcp-parameter-request-list,8A,8C);
}

# global parameters
authoritative;
option nfs-server "%(walt_server_ip)s";
option broadcast-address %(subnet_broadcast)s;

# walt unregistered devices
subnet %(subnet_ip)s netmask %(subnet_netmask)s {
    # declare ranges of unallocated addresses
%(walt_unallocated_ranges_conf)s

    # load default image, if node
    option nfs-fs-path "%(default_fs_path)s";

    # check if the ip is already used
    ping-check = 1;

    # recheck often if you are now registered
    max-lease-time 30;
    default-lease-time 30;
}

# walt registered devices
%(walt_registered_devices_conf)s
"""

RANGE_CONF_PATTERN = "    range %(first)s %(last)s;"

NODE_CONF_PATTERN = """\
host %(hostname)s {
    hardware ethernet %(mac)s;
    fixed-address %(ip)s;
    option host-name "%(hostname)s";
    option nfs-fs-path "%(fs_path)s";
}
"""

SWITCH_CONF_PATTERN = """\
host %(hostname)s {
    hardware ethernet %(mac)s;
    fixed-address %(ip)s;
}
"""

# see http://stackoverflow.com/questions/2154249/identify-groups-of-continuous-numbers-in-a-list
def get_contiguous_ranges(ips):
    ranges=[]
    for k, g in groupby(enumerate(ips), lambda (i,x):i-int(x)):
        group = map(itemgetter(1), g)
        ranges.append((group[0], group[-1]))
    return ranges

def generate_dhcpd_conf(devices):
    subnet = net(const.WALT_SUBNET)
    devices_confs = []
    free_ips = list(subnet.hosts())
    server_ip = free_ips.pop(0)
    for device_info in devices:
        if device_info['type'] == 'switch':
            conf_pattern = SWITCH_CONF_PATTERN
        else:
            conf_pattern = NODE_CONF_PATTERN
        devices_confs.append(conf_pattern % device_info)
        free_ips.remove(ip(device_info['ip']))
    range_confs = []
    for r in get_contiguous_ranges(free_ips):
        first, last = r
        range_confs.append(\
            RANGE_CONF_PATTERN % dict(
                    first=first,
                    last=last
        ))
    default_fs = image.get_mount_path(const.DEFAULT_IMAGE)
    infos = dict(
        default_fs_path=default_fs,
        walt_server_ip=server_ip,
        subnet_ip=subnet.network_address,
        subnet_broadcast=subnet.broadcast_address,
        subnet_netmask=subnet.netmask,
        walt_registered_devices_conf='\n'.join(devices_confs),
        walt_unallocated_ranges_conf='\n'.join(range_confs)
    )
    return CONF_PATTERN % infos

QUERY_DEVICES_WITH_IP="""
    SELECT * FROM devices WHERE ip IS NOT NULL ORDER BY mac;
"""

class DHCPServer(object):
    def __init__(self, db):
        self.db = db
    def update(self, force=False):
        devices = []
        for item in \
                self.db.execute(QUERY_DEVICES_WITH_IP).fetchall():
            device_type = item.type
            device_mac = item.mac
            fs_path = None  # default
            if is_a_node_type_name(device_type):
                image_name = self.db.select_unique(
                            'nodes', mac=device_mac).image
                fs_path = image.get_mount_path(image_name)
            if device_type != 'server':
                devices.append(dict(
                    type=item.type,
                    hostname=item.name,
                    ip=item.ip,
                    mac=device_mac,
                    fs_path=fs_path))
        conf = generate_dhcpd_conf(devices)
        with open(DHCPD_CONF_FILE, 'r') as conf_file:
            old_conf = conf_file.read()
        if conf != old_conf:
            with open(DHCPD_CONF_FILE, 'w') as conf_file:
                conf_file.write(conf)
            force = True # perform the restart below
        if force == True:
            do('service isc-dhcp-server restart')
            print 'dhcpd conf updated.'




#!/usr/bin/env python

from walt.common.tools import get_mac_address
from walt.common.nodetypes import get_node_type_from_mac_address
from walt.common.nodetypes import is_a_node_type_name
from walt.server.network.tools import ip_in_walt_network, lldp_update
import snmp, const, time, re
from tree import Tree
import walt.server

DEVICE_NAME_NOT_FOUND="""No device with name '%s' found.\n"""

TOPOLOGY_QUERY = """
    SELECT  d1.name as name, d1.type as type, d1.mac as mac,
            d1.ip as ip, d1.reachable as reachable,
            d2.name as switch_name, t.switch_port as switch_port
    FROM devices d1, devices d2, topology t
    WHERE   d1.mac = t.mac and d2.mac = t.switch_mac
    UNION ALL
    SELECT  d1.name as name, d1.type as type, d1.mac as mac,
            d1.ip as ip, d1.reachable as reachable,
            NULL as switch_name, NULL as switch_port
    FROM devices d1, topology t
    WHERE   d1.mac = t.mac and t.switch_mac is null
    ORDER BY switch_name, switch_port;"""

# the split_part() expression below allows to show only
# the image tag to the user (instead of the full docker name).
NODE_LIST_QUERY = """
    SELECT  d.name as name, d.type as type,
            split_part(n.image, ':', 2) as image,
            d.ip as ip, d.reachable as reachable
    FROM devices d, nodes n
    WHERE   d.mac = n.mac; """

NEW_NAME_ERROR_AND_GUIDELINES = """\
Failed: invalid new name.
This name must be a valid network hostname. 
Only alphanumeric characters and '-' are allowed.
The 1st character must be an alphabet letter.
(Example: rpi-D327)
"""

class Topology(object):

    def __init__(self, db):
        self.db = db

    def get_type(self, mac):
        node_type = get_node_type_from_mac_address(mac)
        if node_type != None:
            # this is a node
            return node_type.SHORT_NAME
        elif mac == self.server_mac:
            return 'server'
        else:
            return 'switch'

    def collect_connected_devices(self, host, host_is_a_switch,
                            host_mac, processed_switches):

        print "collect devices connected on %s" % host
        # avoid to loop forever...
        if host_is_a_switch:
            processed_switches.add(host_mac)
        while True:
            issue = False
            # get a SNMP proxy with LLDP feature
            snmp_proxy = snmp.Proxy(host, lldp=True)

            # record neighbors in db and recurse
            for port, neighbor_info in snmp_proxy.lldp.get_neighbors().items():
                ip, mac = neighbor_info['ip'], neighbor_info['mac']
                if mac in processed_switches:
                    continue
                if not ip_in_walt_network(ip):
                    print 'Not ready, one neighbor has ip %s (not in WalT network yet)...' % ip
                    lldp_update()
                    time.sleep(1)
                    issue = True
                    break
                device_type = self.get_type(mac)
                if host_is_a_switch:
                    switch_mac, switch_port = host_mac, port
                else:
                    switch_mac, switch_port = None, None
                self.add_device(type=device_type,
                                mac=mac,
                                switch_mac=switch_mac,
                                switch_port=switch_port,
                                ip=ip)
                if device_type == 'switch':
                    # recursively discover devices connected to this switch
                    self.collect_connected_devices(ip, True,
                                            mac, processed_switches)
            if not issue:
                break   # otherwise restart the loop

    def update(self, requester=None):
        # delete some information that will be updated
        self.db.execute('DELETE FROM topology;')
        self.db.execute("""
            UPDATE devices
            SET reachable = 0;""")

        self.server_mac = get_mac_address(const.SERVER_TESTBED_INTERFACE)
        self.collect_connected_devices("localhost", False, self.server_mac, set())
        self.db.commit()

        if requester != None:
            requester.stdout.write('done.\n')

    def get_device_info(self, requester, device_name, \
                        err_message = DEVICE_NAME_NOT_FOUND):
        device_info = self.db.select_unique("devices", name=device_name)
        if device_info == None and err_message != None:
            requester.stderr.write(err_message % device_name)
        return device_info

    def get_node_info(self, requester, node_name):
        node_info = self.get_device_info(requester, node_name)
        if node_info == None:
            return None # error already reported
        device_type = node_info.type
        if not is_a_node_type_name(device_type):
            requester.stderr.write('%s is not a node, it is a %s.\n' % \
                                    (node_name, device_type))
            return None
        return node_info

    def get_reachable_node_info(self, requester, node_name, after_rescan = False):
        node_info = self.get_node_info(requester, node_name)
        if node_info == None:
            return None # error already reported
        if node_info.reachable == 0:
            if after_rescan:
                requester.stderr.write(
                        'Connot reach %s. The node seems dead or disconnected.\n' % \
                                    node_name)
                return None
            else:
                # rescan, just in case, and retry
                self.update()   # rescan, just in case
                return self.get_reachable_node_info(
                        requester, node_name, after_rescan = True)
        return node_info

    def notify_unknown_ip(self, requester, device_name):
        requester.stderr.write('Sorry, IP address of %s in unknown.\n' \
                            % device_name)

    def get_device_ip(self, requester, device_name):
        device_info = self.get_device_info(requester, device_name)
        if device_info == None:
            return None # error already reported
        if device_info.ip == None:
            self.notify_unknown_ip(requester, device_name)
        return device_info.ip

    def get_node_ip(self, requester, node_name):
        node_info = self.get_node_info(requester, node_name)
        if node_info == None:
            return None # error already reported
        if node_info.ip == None:
            self.notify_unknown_ip(requester, node_name)
        return node_info.ip

    def get_reachable_node_ip(self, requester, node_name):
        node_info = self.get_reachable_node_info(requester, node_name)
        if node_info == None:
            return None # error already reported
        return node_info.ip

    def get_connectivity_info(self, requester, node_name):
        node_info = self.get_reachable_node_info(requester, node_name)
        if node_info == None:
            return None # error already reported
        node_mac = node_info.mac
        topology_info = self.db.select_unique("topology", mac=node_mac)
        switch_mac = topology_info.switch_mac
        switch_port = topology_info.switch_port
        switch_info = self.db.select_unique("devices", mac=switch_mac)
        return dict(
            switch_ip = switch_info.ip,
            switch_port = switch_port
        )

    def rename_device(self, requester, old_name, new_name):
        device_info = self.get_device_info(requester, old_name)
        if device_info == None:
            return
        if self.get_device_info(requester, new_name, err_message = None) != None:
            requester.stderr.write("""Failed: a device with name '%s' already exists.\n""" % new_name)
            return
        if not re.match('^[a-zA-Z][a-zA-Z0-9-]*$', new_name):
            requester.stderr.write(NEW_NAME_ERROR_AND_GUIDELINES)
            return
        # all is fine, let's update it
        self.db.update("devices", 'mac', mac = device_info.mac, name = new_name)
        self.db.commit()

    def generate_device_name(self, **kwargs):
        if kwargs['type'] == 'server':
            return 'walt-server'
        return "%s-%s" %(
            kwargs['type'],
            "".join(kwargs['mac'].split(':')[3:]))

    def add_device(self, **kwargs):
        # if we are there then we can reach this device
        kwargs['reachable'] = 1
        # update device info
        num_rows = self.db.update("devices", 'mac', **kwargs)
        # if no row was updated, this is a new device
        if num_rows == 0:
            # generate a name for this device
            name = self.generate_device_name(**kwargs)
            kwargs['name'] = name
            # insert a new row
            self.db.insert("devices", **kwargs)
            # add node info if relevant
            if is_a_node_type_name(kwargs['type']):
                # unknown nodes boot with the default image
                default_image = walt.server.instance.images.get_default_image()
                self.db.insert("nodes", image=default_image, **kwargs)
        # add topology info
        self.db.insert("topology", **kwargs)

    def printed_as_tree(self):
        t = Tree()
        for device in self.db.execute(TOPOLOGY_QUERY).fetchall():
            name = device.name
            swport = device.switch_port
            if swport == None:
                label = name
                # align to 2nd letter of the name
                subtree_offset = 1
            else:
                label = '%d: %s' % (swport, name)
                # align to 2nd letter of the name
                subtree_offset = label.find(' ') + 2
            parent_key = device.switch_name
            t.add_node( name,   # will be the key in the tree
                        label,
                        subtree_offset=subtree_offset,
                        parent_key = parent_key)
        return t.printed()

    def printed_as_detailed_table(self):
        return self.db.pretty_printed_select(
                    TOPOLOGY_QUERY)

    def list_nodes(self):
        return self.db.pretty_printed_select(
                    NODE_LIST_QUERY)

    def register_node(self, node_ip):
        row = self.db.select_unique("devices", ip=node_ip)
        if row == None or row.reachable == 0:
            print 'New node detected.'
            # restart discovery
            self.update()
            # update dhcpd
            walt.server.instance.dhcpd.update()

